@ Copyright (c) 2003, Torlus <greg@heliscar.com>
@ see LICENCE.txt for further details

  .GLOBAL Fill
  .GLOBAL Perspective
  
	.extern DIVTAB

	.section	.iwram,"ax",%progbits
	.arm
	.align
	.TEXT

@void Fill(void *addr, s16 val, s32 count)
@ Fills memory by "lines" of 240 16-bit values
@ suitable for ScreenBuffer and ZBuffer
Fill:
  stmfd	sp!, {v1-v7,fp,ip,lr}            @ push registers we will corrupt

  @ convert val to 32-bit value
  add a2, a2, a2, lsl #16

  mov a4, a2
  mov v1, a4
  mov v2, v1
  mov v3, v2
  mov v4, v3
  mov v5, v4
  mov v6, v5
  mov v7, v6
  mov fp, v7
  mov ip, fp
  mov lr, fp

Fill_loop:

  @ clears 9600 bytes

  .rept 200               @ unroll loop
  stmia a1!, {a2,a4,v1-v7,fp,ip,lr} @ 48 bytes
  .endr

  subs  a3, a3, #1        @ decrease line counter
  bne   Fill_loop

  ldmfd	sp!, {v1-v7,fp,ip,lr}            @ push registers we will corrupt
  bx    lr


@void Perspective(vertex *first_vertex,u32 count)
@ Performs 3D to 2D conversion on a set of vertices
Perspective:

@ v7 will walk throuh memory
@ v6 is the vertex counter
@ v5 points to division table

  stmfd	sp!, {v4-v7}            @ push registers we will corrupt

  add   v7, a1, #(4*4)          @ make v7 point to first_vertex and skip bone & rel[] in vertex structure
  mov   v6, a2                  @ v6 is vertex counter

  ldr   v5, =DIVTAB
  add   v5, v5, #(ZCAM*4)       @ v5 points on DIVTAB + ZCAM (* 4 bytes)
  
Perspective_loop:

  ldr   a4, [v7], #4            @ a4 = x
  mov   a4, a4, asr #8          @ a4 = fp2int(x)

  ldr   a3, [v7], #4            @ a3 = y
  mov   a3, a3, asr #8          @ a3 = fp2int(y)

  ldr   v4, [v7], #4            @ v4 = z
  mov   v4, v4, asr #8          @ v4 = fp2int(z)

  eor   a3, a3, v4              
  eor   v4, v4, a3
  eor   a3, a3, v4
  rsb   a3, a3, #0

  ldr   v4, [v5, v4, lsl #2]    @ v4 = divtab[z+ZCAM]
.ifdef GBA3DS_MODE5
  mul   v3, a4, v4              @ v3 = fp2int(x) * divtab[z+ZCAM]
  mov   a4, v3, asr #(SCALE_SX_ASR)  @ scale it
  add   a4, a4, #(HEIGHT/2)   @ a4 = Sx (int)

  mul   v3, a3, v4              @ v3 = fp2int(y) * divtab[z+ZCAM]
  mov   a3, v3, asr #(SCALE_SY_ASR)  @ scale it
  add   a3, a3, #(WIDTH/2)  @ a3 = Sy (int)

  eor   a3, a3, a4
  eor   a4, a4, a3
  eor   a3, a3, a4
  
.else
  mul   v3, a4, v4              @ v3 = fp2int(x) * divtab[z+ZCAM]
  mov   a4, v3, asr #(SCALE_SX_ASR)  @ scale it
  add   a4, a4, #(WIDTH/2)   @ a4 = Sx (int)

  mul   v3, a3, v4              @ v3 = fp2int(y) * divtab[z+ZCAM]
  mov   a3, v3, asr #(SCALE_SY_ASR)  @ scale it
  add   a3, a3, #(HEIGHT/2)  @ a3 = Sy (int)
.endif

  str   a3, [v7], #4            @ stores Sy
  str   a4, [v7], #(7*4)        @ stores Sx and moves to next vertex
                                @ (skips Sx,u,v,bone,rel[])

  sub   v6, v6, #1                  @ check if we have vertices remaining
  cmp   v6, #0
  bgt   Perspective_loop


  ldmfd	sp!, {v4-v7}		        @ pop registers
  bx    lr                      @ and exit
        
	.pool

	.END

