// Copyright (c) 2003, Torlus <greg@heliscar.com>
// see LICENCE.txt for further details

#include "matrix.h"
#include "trig.h"
#include "gbautil.h"

void IWRAM matrix_loadIdentity(s32 *mat) {
  int i;
  for(i=0; i<16;i++) {
    mat[i] = 0;
  }
  mat[0] = mat[5] = mat[10] = mat[15] = (1<<8);
}


void IWRAM matrix_copy(s32 *src, s32 *dst) {
  int i;
  for(i=0; i<16;i++) {
    dst[i] = src[i];
  }
}

void IWRAM matrix_postMultiply(s32 *m1, s32 *m2, s32 *dst) {

  s32 tmp[16];

  tmp[ 0 ] = (m1[ 0 ] * m2[ 0 ] + m1[ 4 ] * m2[ 1 ] + m1[ 8 ] * m2[ 2 ])>>8;
  tmp[ 1 ] = (m1[ 1 ] * m2[ 0 ] + m1[ 5 ] * m2[ 1 ] + m1[ 9 ] * m2[ 2 ])>>8;
  tmp[ 2 ] = (m1[ 2 ] * m2[ 0 ] + m1[ 6 ] * m2[ 1 ] + m1[ 10 ] * m2[ 2 ])>>8;
  tmp[ 3 ] = 0;

  tmp[ 4 ] = (m1[ 0 ] * m2[ 4 ] + m1[ 4 ] * m2[ 5 ] + m1[ 8 ] * m2[ 6 ])>>8;
  tmp[ 5 ] = (m1[ 1 ] * m2[ 4 ] + m1[ 5 ] * m2[ 5 ] + m1[ 9 ] * m2[ 6 ])>>8;
  tmp[ 6 ] = (m1[ 2 ] * m2[ 4 ] + m1[ 6 ] * m2[ 5 ] + m1[ 10 ] * m2[ 6 ])>>8;
  tmp[ 7 ] = 0;

  tmp[ 8 ] = (m1[ 0 ] * m2[ 8 ] + m1[ 4 ] * m2[ 9 ] + m1[ 8 ] * m2[ 10 ])>>8;
  tmp[ 9 ] = (m1[ 1 ] * m2[ 8 ] + m1[ 5 ] * m2[ 9 ] + m1[ 9 ] * m2[ 10 ])>>8;
  tmp[ 10 ] = (m1[ 2 ] * m2[ 8 ] + m1[ 6 ] * m2[ 9 ] + m1[ 10 ] * m2[ 10 ])>>8;
  tmp[ 11 ] = 0;

  tmp[ 12 ] = (m1[ 0 ] * m2[ 12 ] + m1[ 4 ] * m2[ 13 ] + m1[ 8 ] * m2[ 14 ] + m1[ 12 ])>>8;
  tmp[ 13 ] = (m1[ 1 ] * m2[ 12 ] + m1[ 5 ] * m2[ 13 ] + m1[ 9 ] * m2[ 14 ] + m1[ 13 ])>>8;
  tmp[ 14 ] = (m1[ 2 ] * m2[ 12 ] + m1[ 6 ] * m2[ 13 ] + m1[ 10 ] * m2[ 14 ] + m1[ 14 ])>>8;
  tmp[ 15 ] = (1<<8);

  matrix_copy( tmp, dst );
}

void IWRAM matrix_setTranslation( s32 *mat, s32 *trs ) {
  mat[ 12 ] = trs[ 0 ];
  mat[ 13 ] = trs[ 1 ];
  mat[ 14 ] = trs[ 2 ];
}

void IWRAM matrix_setInverseTranslation( s32 *mat, s32 *trs ) {
  mat[ 12 ] = -trs[ 0 ];
  mat[ 13 ] = -trs[ 1 ];
  mat[ 14 ] = -trs[ 2 ];
}

void IWRAM matrix_setRotation( s32 *mat, s32 *ang ) {
  s32 cr = COS[ ang[ 0 ] ];
  s32 sr = SIN[ ang[ 0 ] ];
  s32 cp = COS[ ang[ 1 ] ];
  s32 sp = SIN[ ang[ 1 ] ];
  s32 cy = COS[ ang[ 2 ] ];
  s32 sy = SIN[ ang[ 2 ] ];

  s32 srsp = (sr * sp)>>8;
  s32 crsp = (cr * sp)>>8;

  mat[ 0 ] = (cp * cy)>>8;
  mat[ 1 ] = (cp * sy)>>8;
  mat[ 2 ] = (-sp);

  mat[ 4 ] = ( srsp * cy - cr * sy )>>8;
  mat[ 5 ] = ( srsp * sy + cr * cy )>>8;
  mat[ 6 ] = ( sr * cp )>>8;

  mat[ 8 ] = ( crsp * cy + sr * sy )>>8;
  mat[ 9 ] = ( crsp * sy - sr * cy )>>8;
  mat[ 10 ] = ( cr * cp )>>8;
}

void IWRAM matrix_setInverseRotation( s32 *mat, s32 *ang ) {
  s32 cr = COS[ ang[ 0 ] ];
  s32 sr = SIN[ ang[ 0 ] ];
  s32 cp = COS[ ang[ 1 ] ];
  s32 sp = SIN[ ang[ 1 ] ];
  s32 cy = COS[ ang[ 2 ] ];
  s32 sy = SIN[ ang[ 2 ] ];

  s32 srsp = (sr * sp)>>8;
  s32 crsp = (cr * sp)>>8;

  mat[ 0 ] = (cp * cy)>>8;
  mat[ 4 ] = (cp * sy)>>8;
  mat[ 8 ] = (-sp);

  mat[ 1 ] = ( srsp * cy - cr * sy )>>8;
  mat[ 5 ] = ( srsp * sy + cr * cy )>>8;
  mat[ 9 ] = ( sr * cp )>>8;

  mat[ 2 ] = ( crsp * cy + sr * sy )>>8;
  mat[ 6 ] = ( crsp * sy - sr * cy )>>8;
  mat[ 10 ] = ( cr * cp )>>8;
}

void IWRAM matrix_inverseRotateVect( s32 *mat, s32 *vec, s32 *vecdst ) {
    s32 tmp[3];

    tmp[0] = (vec[0] * mat[0] + vec[1] * mat[1] + vec[2] * mat[2])>>8;
    tmp[1] = (vec[0] * mat[4] + vec[1] * mat[5] + vec[2] * mat[6])>>8;
    tmp[2] = (vec[0] * mat[8] + vec[1] * mat[9] + vec[2] * mat[10])>>8;

    vecdst[0] = tmp[0];
    vecdst[1] = tmp[1];
    vecdst[2] = tmp[2];
}

void IWRAM matrix_inverseTranslateVect( s32 *mat, s32 *vec, s32 *vecdst ) {

    vecdst[0] = vec[0]-mat[12];
    vecdst[1] = vec[1]-mat[13];
    vecdst[2] = vec[2]-mat[14];

}

void IWRAM matrix_transformVertex( s32 *mat, s32 *vec, s32 *vecdst ) {
    s32 tmp[ 3 ];

    tmp[ 0 ] = ((vec[0] * mat[0] + vec[1] * mat[4] + vec[2] * mat[8])>>8) + mat[12];
    tmp[ 1 ] = ((vec[0] * mat[1] + vec[1] * mat[5] + vec[2] * mat[9])>>8) + mat[13];
    tmp[ 2 ] = ((vec[0] * mat[2] + vec[1] * mat[6] + vec[2] * mat[10])>>8) + mat[14];

    vecdst[0] = tmp[0];
    vecdst[1] = tmp[1];
    vecdst[2] = tmp[2];
}
