@ Copyright (c) 2003, Torlus <greg@heliscar.com>
@ see LICENCE.txt for further details

@ Contributions of : subbie

  .GLOBAL Draw

	.ifdef GBA3DS_MODE3
		.extern ScreenBuffer
	.else
  		.extern VideoBuffer
	.endif
	.extern DIVTAB
	.extern EdgeLeft
	.extern EdgeRight
  
	.section	.iwram,"ax",%progbits
	.arm
	.align
	.TEXT

@void Draw(triangle *tri, vertex *first_vertex, u16 **texture, u32 count)
Draw:
  stmfd   sp!, {v1-v7,fp,ip,lr}

  add     a2, a2, #(5*4) @ a2 points on vertices (and skips bone, rel[],abs[0])

Draw_Loop:		@ loops through triangles
  bl    DrawFace

  add   a1, a1, #(4*2)

  sub   a4, a4, #1
  cmp   a4, #0
  bgt   Draw_Loop

  ldmfd   sp!, {v1-v7,fp,ip,lr}
  bx    lr

DrawFace:
  stmfd   sp!, {a1-a4,lr}

  ldrh    fp, [a1], #2
  ldr     fp, [a3, fp, lsl #2]
  
  mov     v7, a2

  ldr     lr, =(11*4)   @ sizeof(vertex)

  ldrh     v5, [a1], #2  @ reads point 1 index
  mla     a2, v5, lr, v7  @ a2 points to "y" of point 1

  ldrh     v5, [a1], #2  @ same for point 2, in a3
  mla     a3, v5, lr, v7

  ldrh     v5, [a1], #2  @ same for point 3, in a4
  mla     a4, v5, lr, v7 

  ldr     v2, [a2]  @ v2 = y point 1
  cmp     v2, #(ZMIN<<8)
  blt     DF_End
  ldr     v3, [a3]  @ v3 = y point 2
  cmp     v3, #(ZMIN<<8)
  blt     DF_End
  ldr     v4, [a4]  @ v4 = y point 3
  cmp     v4, #(ZMIN<<8)
  blt     DF_End

  @ Backface culling : for 2 given vectors, check the sign of 3rd coordinate of the cross product
  @ if negative, the face is not visible

  ldr     v2, [a2, #12]   @ v2 = Sx point 1
  ldr     v3, [a3, #12]   @ v3 = Sx point 2
  ldr     v4, [a4, #12]   @ v4 = Sx point 3

  sub     ip, v3, v2     @ ip = Sx12 (x1)
  sub     lr, v4, v2     @ lr = Sx13 (x2)

  ldr     v2, [a2, #8]   @ v2 = Sy point 1
  ldr     v3, [a3, #8]   @ v3 = Sy point 2
  ldr     v4, [a4, #8]   @ v4 = Sy point 3

  sub     a1, v3, v2     @ a1 = Sy12 (y1)
  sub     v5, v4, v2     @ v5 = Sy13 (y2)

  mul     v6, ip, v5
  mul     v7, lr, a1
  cmp     v6, v7
.ifdef GBA3DS_MODE5
  ble	  DF_End
.else
  bge     DF_End
.endif

  @ sort according to Sy
  @ these "eor" perform register swap

  cmp     v4, v3
  eorlt   v3, v3, v4
  eorlt   v4, v4, v3
  eorlt   v3, v3, v4
  eorlt   a3, a3, a4
  eorlt   a4, a4, a3
  eorlt   a3, a3, a4

  cmp     v3, v2
  eorlt   v2, v2, v3
  eorlt   v3, v3, v2
  eorlt   v2, v2, v3
  eorlt   a2, a2, a3
  eorlt   a3, a3, a2
  eorlt   a2, a2, a3

  cmp     v4, v3
  eorlt   v3, v3, v4
  eorlt   v4, v4, v3
  eorlt   v3, v3, v4
  eorlt   a3, a3, a4
  eorlt   a4, a4, a3
  eorlt   a3, a3, a4

  cmp     v2, v4        @ yMin == yMax ?
  beq     DF_End

  cmp     v2, #(HEIGHT-1) @ yMin >= HEIGHT ?
  bge     DF_End

  cmp     v4, #0        @ yMax <= 0 ?
  ble     DF_End

  mov     v6, v2		@ stores yMin and yMax somewhere
  mov     v7, v4

  @ now we check how is made the triangle
  @ it may have two edges on its left and one on its right
  @ or two edges on its right and one on its left

  @ then we will fill the EdgeLeft and EdgeRight arrays

  ldr     v3, [a3, #12]   @ v3 = Sx point 2
  ldr     v4, [a4, #12]   @ v4 = Sx point 3

  cmp     v3, v4
  blt     DF_Edge_LL_R
  bgt     DF_Edge_L_RR

  ldr     v2, [a2, #12]   @ v2 = Sx point 1

  cmp     v2, v3
  blt     DF_Edge_L_RR
  
DF_Edge_LL_R:

  ldr   v5, =EdgeLeft
  mov   v2, a2
  mov   v4, a3
  bl      BuildEdge

  mov   v2, a3
  mov   v4, a4

  bl      BuildEdge

  ldr   v5, =EdgeRight
  mov   v2, a2
  mov   v4, a4


  bl      BuildEdge

  b       DFEdgeSort

DF_Edge_L_RR:

  ldr   v5, =EdgeLeft
  mov   v2, a2
  mov   v4, a4
  bl      BuildEdge

  ldr   v5, =EdgeRight
  mov   v2, a2
  mov   v4, a3
  bl      BuildEdge

  mov   v2, a3
  mov   v4, a4
  bl      BuildEdge

DFEdgeSort:
  @ this code should be removed. It used it on a previous version
  @ of the engine that could do n-gones
  @ but if I remove it, there are some bugs :)
  @ I'll have a better look at that later

  @ this code basically determine which Edge is on the left, and
  @ which one is on the right.

  mov     a1, v6
  mov     a2, v7

  ldr     v2, =EdgeLeft
  ldr     v4, =EdgeRight

  cmp     a1, #0
  movlt   a1, #0
  cmp     a2, #(HEIGHT-1)
  ldrgt   a2, =(HEIGHT-1)
  add     v2, v2, a1, lsl #4
  add     v4, v4, a1, lsl #4

  

DFEdgeSort_loop:

  ldr     a3, [v2]    @ Sx1 (EdgeLeft)
  ldr     a4, [v4]    @ Sx2 (EdgeRight)

  cmp     a3, a4        @ loop until the Sx1 and Sx2 values are equal 
  addeq   v2, v2, #16
  addeq   v4, v4, #16
  addeq   a1, a1, #1
  beq     DFEdgeSort_loop

  eorgt   v2, v2, v4    @ if Sx1 > Sx2, swap EdgeLeft and EdgeRight
  eorgt   v4, v4, v2
  eorgt   v2, v2, v4

DFDrawLines:

  @ intialize ZBuffer and ScreenBuffer pointers according to Ymin (y1)

  ldr     v6, =ZBuffer
.ifdef GBA3DS_MODE5
  add 	  v6, v6, a1, lsl #8
  add	  v6, v6, a1, lsl #6   @ adds (y1*160*2)
.else
  add     v6, v6, a1, lsl #9
  sub     v6, v6, a1, lsl #5   @ adds (y1*240*2)
.endif


.ifdef GBA3DS_MODE3
  ldr     v7, =ScreenBuffer
  add     v7, v7, a1, lsl #9
  sub     v7, v7, a1, lsl #5   @ adds (y1*240*2)
.else
  ldr     v7, =VideoBuffer
  ldr     v7, [v7]
.ifdef GBA3DS_MODE4
  add     v7, v7, a1, lsl #8
  sub     v7, v7, a1, lsl #4   @ adds (y1*120*2)
.else
  add 	  v7, v7, a1, lsl #8
  add	  v7, v7, a1, lsl #6   @ adds (y1*160*2)
.endif
.endif

  ldrh    v5, [fp], #4        @ width of texture
  add     v5, v5, #1  		  @ 16 bits-wide

  sub     a2, a2, a1

DFDrawLines_loop:

  bl      ScanLine

  add     v2, v2, #16         @ Edge Down
  add     v4, v4, #16         @ Edge Up
  add     v6, v6, #(WIDTH*2)    @ ZBuffer

.ifdef GBA3DS_MODE4
  add     v7, v7, #(WIDTH)    @ ScreenBuffer
.else
  add     v7, v7, #(WIDTH*2)    @ ScreenBuffer
.endif

  subs   a2, a2, #1
  bpl   DFDrawLines_loop

DF_End:

  ldmfd   sp!, {a1-a4,lr}
  bx      lr




BuildEdge:

@ v2 points on "z" of 1er point
@ v4 points on "z" du 2eme point
@ v5 points on EdgeLeft or EdgeRight
  stmfd   sp!, {a1-a4,v1-v7,fp,lr}

  ldr     a1, [v2], #8     @ reads z1
  ldr     a2, [v2], #4      @ reads Sy1

  ldr     a3, [v4], #8     @ reads z2
  ldr     a4, [v4], #4      @ reads Sy2

  cmp     a2, a4            @ if Sy1 > Sy2 swap 
  beq     BE_End

  eorgt   a2, a2, a4
  eorgt   a4, a4, a2
  eorgt   a2, a2, a4
  eorgt   a1, a1, a3
  eorgt   a3, a3, a1
  eorgt   a1, a1, a3
  eorgt   v2, v2, v4        @ swap pointers 
  eorgt   v4, v4, v2
  eorgt   v2, v2, v4


	@ bounds checking

  cmp     a2, #(HEIGHT-1)
  bgt     BE_End

  cmp     a4, #0
  blt     BE_End

  mov     lr, fp

  ldr     v6, =DIVTAB
  sub     v7, a4, a2

  ldr     v7, [v6, v7, lsl #2]  @ v7 = divtab[dSy]

  ldr     fp, [v2], #4      @ fp = Sx1
  mov     fp, fp, lsl #8    @ int2fp

  @ reads width and height of texture
  ldrh    v6, [lr], #2
  ldrh    lr, [lr]

  ldr     v1, [v2], #4      @ v1 = U1
  mov     v1, v1, lsl v6
  ldr     v2, [v2]          @ v2 = V1
  mov     v2, v2, lsl lr

  ldr     ip, [v4], #4      @ ip = Sx2
  mov     ip, ip, lsl #8    @ int2fp

  ldr     v3, [v4], #4      @ v3 = U2
  mov     v3, v3, lsl v6
  ldr     v4, [v4]          @ v4 = V2
  mov     v4, v4, lsl lr

  @ (z1, Sy1, Sx1) = (a1,a2,fp)   (U1, V1) = (v1, v2)

  @ (z2, Sy2, Sx2) = (a3,a4,ip)   (U2, V2) = (v3, v4)
  @ v5 = Edge

  @ calculates dZ, dX, dU, dV

  sub     a3, a3, a1
  mul     v6, a3, v7
  mov     a3, v6, asr #16

  sub     ip, ip, fp
  mul     v6, ip, v7
  mov     ip, v6, asr #16

  sub     v3, v3, v1
  mul     v6, v3, v7
  mov     v3, v6, asr #16

  sub     v4, v4, v2
  mul     v6, v4, v7
  mov     v4, v6, asr #16

  @ if Sy1 < 0, adjust

  cmp   a2, #0
  rsblt a2, a2, #0
  mlalt a1, a3, a2, a1
  mlalt fp, ip, a2, fp
  mlalt v1, v3, a2, v1
  mlalt v2, v4, a2, v2
  movlt a2, #0

  cmp   a4, #(HEIGHT-1)
  ldrgt a4, =(HEIGHT-1)

  sub   a4, a4, a2

  add   v5, v5, a2, lsl #4        @ point to right place on Edge

BE_loop:	@fills the structure

  str   fp, [v5], #4
  str   a1, [v5], #4
  str   v1, [v5], #4
  str   v2, [v5], #4

  add   fp, fp, ip
  add   a1, a1, a3
  add   v1, v1, v3
  add   v2, v2, v4

  subs   a4, a4, #1
  bpl   BE_loop

BE_End:

  ldmfd   sp!, {a1-a4,v1-v7,fp,lr}
  bx      lr




ScanLine:
@ v6 points on a ZBuffer "line"
@ v2 points on EdgeLeft
@ v4 points on EdgeRight
@ fp points on texture

  stmfd   sp!, {a1-a4,v1-v7,fp,lr}

  mov     lr, fp      @ lr = texture

  ldr     a2, [v2], #4      @ Sx1
  ldr     a1, [v2], #4      @ z1

  ldr     a4, [v4], #4      @ Sx2
  ldr     a3, [v4], #4      @ z2

  mov     a2, a2, asr #8    @ Sx1 and Sx2 to int values
  mov     a4, a4, asr #8

	@ bound checking
  cmp     a2, #(WIDTH-1)
  bgt     SL_End

  cmp     a4, #0
  blt     SL_End

  ldr     fp, =DIVTAB

  sub     ip, a4, a2

  ldr     ip, [fp, ip, lsl #2]  @ ip = divtab[dY]

  ldr     v1, [v2], #4      @ v1 = U1
  ldr     v2, [v2]          @ v2 = V1

  ldr     v3, [v4], #4      @ v3 = U2
  ldr     v4, [v4]          @ v4 = V2

  @ calculates dZ, dU, dV

  sub     a3, a3, a1
  mul     fp, a3, ip
  mov     a3, fp, asr #16

  sub     v3, v3, v1
  mul     fp, v3, ip
  mov     v3, fp, asr #16

  sub     v4, v4, v2
  mul     fp, v4, ip
  mov     v4, fp, asr #16

  @ adjust if Sx1 < 0

  cmp   a2, #0
  rsblt a2, a2, #0
  mlalt a1, a3, a2, a1
  mlalt v1, v3, a2, v1
  mlalt v2, v4, a2, v2
  movlt a2, #0

  cmp   a2, #(WIDTH-1)
  bgt   SL_End

  cmp   a4, #(WIDTH-1)
  ldrgt a4, =(WIDTH-1)

.ifdef GBA3DS_MODE4
  mov   a2, a2, asr #1
  mov   a4, a4, asr #1
.endif

  sub   a4, a4, a2

  add   v6, v6, a2, lsl #1        @ advance in ZBuffer
  add   v7, v7, a2, lsl #1        @ advance in ScreenBuffer

SL_loop:

  ldrh  fp, [v6]

  mov   ip, a1, asr #8            @ Z to int
  cmp   fp, ip
  addlt v6, v6, #2
  addlt v7, v7, #2
  blt   SL_loop_next              @ test ZBuffer

  strh  ip, [v6], #2

@ texture V coordinate
  mov   fp, v2, asr #8      @ V to int
  add   ip, lr, fp, lsl v5
@ texture U coordinate
  mov   fp, v1, asr #8      @ U en int
  add ip, ip, fp, lsl #1  @ 2 bytes

.ifdef GBA3DS_MODE4
  ldrb  ip, [ip]
  add   v1, v1, v3
  add   v2, v2, v4
@ texture V coordinate
  mov   fp, v2, asr #8      @ V to int
  add   a2, lr, fp, lsl v5
@ texture U coordinate
  mov   fp, v1, asr #8      @ U en int
  add a2, a2, fp, lsl #1  @ 2 bytes
  ldrb  a2, [a2]
  add   ip, ip, a2, lsl #8
  strh  ip, [v7], #2
.else
  ldrh  ip, [ip]
@ store colour
  strh  ip, [v7], #2
.endif

SL_loop_next:

  add   a1, a1, a3
  add   v1, v1, v3
  add   v2, v2, v4

  subs   a4, a4, #1
  bpl   SL_loop

SL_End:

  ldmfd   sp!, {a1-a4,v1-v7,fp,lr}
  bx      lr
  
	.pool


	.END

